import multiprocessing
import os
import signal

from multiprocessing import Pipe
from threading import Thread


class TaskRunner:
    def __init__(self, function, timeout=None):
        super(TaskRunner, self).__init__()
        self._function = function
        self._timeout = timeout
        self._terminator_thread = None
        self._terminator_pipe = None

    def build_terminator(self):
        self._terminator_pipe = Pipe(False)
        self._terminator_thread = TerminatorThread(self._terminator_pipe[0])
        return Terminator(self._terminator_pipe[1])

    def run(self, *args, **kwargs):
        receiver, sender = Pipe(False)
        process = multiprocessing.Process(target=TaskRunner._process_task, args=(self._function, sender, args), kwargs=kwargs)
        process.start()
        if self._terminator_pipe is not None:
            self._terminator_thread.start(sender, process.pid)
        process.join(timeout=self._timeout)
        return TaskRunner._receive_response(receiver)

    @staticmethod
    def _process_task(function, sender, args):
        try:
            result = function(*args)
            sender.send(result)
        except Exception as e:
            sender.send(e)

    @staticmethod
    def _receive_response(receiver):
        if receiver.poll():
            result = receiver.recv()
            if isinstance(result, Exception):
                raise result
            else:
                return result


class Terminator:
    def __init__(self, sender):
        self._sender = sender

    def terminate(self, message):
        self._sender.send(message)


class TerminatorThread(Thread):

    def __init__(self, receiver):
        super().__init__(target=self.listen, daemon=True)
        self.receiver = receiver
        self.sender = None
        self.pid = None
        self.pid_alive = True

    def start(self, sender=None, pid=None):
        assert sender is not None
        assert pid is not None
        super().start()
        self.sender = sender
        self.pid = pid

    def listen(self):
        if self.pid_alive:
            if self.receiver.poll(60):
                os.kill(self.pid, signal.SIGTERM)
                self.pid_alive = False
                self.sender.send(TaskTerminatedError(self.receiver.recv()))
            else:
                self.listen()


class TaskTerminatedError(Exception):
    pass
